#include <zephyr/zephyr.h>
#include <stdio.h>
#include <zephyr/drivers/uart.h>
#include <string.h>
#include <nrfx_pdm.h>
#include <hal/nrf_gpio.h>
#include "nrf5340_application.h"
#include <zephyr/kernel.h>
#include <zephyr/logging/log.h>
#include <math.h>

// for storing data
#include <zephyr/storage/disk_access.h>
#include <zephyr/fs/fs.h>
#include <ff.h>
#include <zephyr/kernel.h>
#include <zephyr/device.h>

LOG_MODULE_REGISTER(mic, LOG_LEVEL_DBG);

#define RMS_BUFFER_SIZE 4096
#define MAX_SOUND_PRESSURE_LEVEL 133 // change for specific mic in dB
#define THRESHOLD_LINEAR_VALUE pow(10, (MAX_SOUND_PRESSURE_LEVEL / 10)) // Convert dB to linear scale

#define PDM_BUF_SIZE 1024
#define CLK_PIN 24 // change for specific pin
#define DIN_PIN 27 // change for specific pin

#define PERIOD_MS 2
#define DECIMATION_FACTOR 80  // Change this factor based on your requirements

int16_t pdm_buf[PDM_BUF_SIZE];
int16_t pcm_buf[PDM_BUF_SIZE / DECIMATION_FACTOR];
int16_t rms_buffer[RMS_BUFFER_SIZE]; 
uint16_t rms_buffer_cursor = 0;

static int lsdir(const char *path);
static int write_file(const char *filename, const int16_t *data);

static FATFS fat_fs;
/* mounting info */
static struct fs_mount_t mp = {
    .type = FS_FATFS,						//file system type
    .fs_data = &fat_fs,						//is the file system data which will be used by fs_mount() API.
};

/*
 *  Note the fatfs library is able to mount only strings inside _VOLUME_STRS
 *  in ffconf.h
 */
static const char *disk_mount_pt = "/SD:";
const char *filename = "/SD:/audio.raw";
const int8_t data[0] = {6};

void func(){
    
    write_file(filename, data);
    
    k_sleep(K_MSEC(3000));
}

double sqrt(double value) {
    int i;
    double sqrt = value / 3;

    if (value <= 0) {
        return 0;
    }

    for (i = 0; i < 6; i++) {
        sqrt = (sqrt + value / sqrt) / 2;
    }

    return sqrt;
}

double calculate_rms(int16_t *buffer, uint16_t num_samples) {
    uint64_t sum = 0;

    for (uint16_t j = 0; j < num_samples; j++) {
        printk("PCM buffer data: %d\r\n", buffer[j]);
        rms_buffer[rms_buffer_cursor] = buffer[j];
        rms_buffer_cursor = (rms_buffer_cursor + 1) % RMS_BUFFER_SIZE;
    }

    for (uint16_t j = 0; j < RMS_BUFFER_SIZE; j++) {
        sum += rms_buffer[j] * rms_buffer[j];
    }
    double tosquirt = sum / RMS_BUFFER_SIZE;
    
    return sqrt(tosquirt); // return sqrt(tosquirt)
}

void pdm_to_pcm(int16_t *pdm_buffer, int16_t *pcm_buffer, uint16_t pdm_size) {
    // int32_t sum = 0;
    // uint16_t pcm_index = 0;

    // for (uint16_t i = 0; i < pdm_size; i++) {
    //     sum += pdm_buffer[i];
    //     if ((i + 1) % DECIMATION_FACTOR == 0) {
    //         pcm_buffer[pcm_index++] = sum / DECIMATION_FACTOR;
    //         sum = 0;
    //     }
    // }
    size_t pcm_size = pdm_size / DECIMATION_FACTOR;
    // Implement your decimation logic here
    // This is just a placeholder example
    for (size_t i = 0; i < pcm_size; i++) {
        pcm_buf[i] = pdm_buffer[i * DECIMATION_FACTOR];
        // LOG_INF("PCM buffer: %d",pcm_buf[i]);
    }
}

void audio_callback(int16_t *pdm_buffer, uint16_t size) {
    uint16_t pcm_size = size / DECIMATION_FACTOR;

    // Convert PDM to PCM
    pdm_to_pcm(pdm_buffer, pcm_buf, size);
    // if (write_file(filename, (const uint8_t *)pcm_buf, pcm_size * sizeof(int16_t)) != 0) {
    //     LOG_ERR("Failed to write PCM data to file");
    // }
    


    // write_file(filename,(const uint8_t *)pcm_buf);

    // Calculate RMS value using the PCM buffer
    double rms = calculate_rms(pcm_buf, pcm_size);
    double rms_dB = 20 * log10(rms);

    // Print RMS value in dB
    // printk("RMS value in dB: %.2f dB\n", rms_dB);
}

ISR_DIRECT_DECLARE(pdm_isr_handler) {
    nrfx_pdm_irq_handler();
    ISR_DIRECT_PM();
    return 1;
}

void nrfx_pdm_event_handler(nrfx_pdm_evt_t const *const p_evt) {
    if (p_evt->buffer_requested) {
        nrfx_pdm_buffer_set(pdm_buf, PDM_BUF_SIZE);
    }
    if (p_evt->buffer_released != 0) {
        audio_callback(pdm_buf, PDM_BUF_SIZE);
        // write_file(filename, pcm_buf);
    }
}

static void pdm_init(void) {
    nrfx_pdm_config_t pdm_config = NRFX_PDM_DEFAULT_CONFIG(CLK_PIN, DIN_PIN);
    nrfx_pdm_init(&pdm_config, nrfx_pdm_event_handler);
}

void sd_card_init(){
    do {
        static const char *disk_pdrv = "SD";
        uint64_t memory_size_mb;
        uint32_t block_count;
        uint32_t block_size;

        if (disk_access_init(disk_pdrv) != 0) {						    //performs disk initialisation
            LOG_ERR("Storage init ERROR!");
            break;
        }

        if (disk_access_ioctl(disk_pdrv,								//Function to get disk parameters
                DISK_IOCTL_GET_SECTOR_COUNT, &block_count)) {
            LOG_ERR("Unable to get sector count");						//DISK_IOCTL_GET_SECTOR_COUNT: Get the number of sectors in the disk
            break;
        }
        LOG_INF("Block count %u", block_count);

        if (disk_access_ioctl(disk_pdrv,
                DISK_IOCTL_GET_SECTOR_SIZE, &block_size)) {				//DISK_IOCTL_GET_SECTOR_SIZE: Get the size of a disk SECTOR in bytes.
            LOG_ERR("Unable to get sector size");
            break;
        }
        printk("Sector size %u\n", block_size);

        memory_size_mb = (uint64_t)block_count * block_size;
        printk("Memory Size(MB) %u\n", (uint32_t)(memory_size_mb >> 20));
    } while (0);

    mp.mnt_point = disk_mount_pt;

    int res = fs_mount(&mp);					//mounts the filesystem on disk

    if (res == FR_OK) {
        printk("Disk mounted.\n");
        lsdir(disk_mount_pt);

        
        
        // const char data[200] = "hjdhfjshdklqhihwbcjsdiwhirfkbekfjqoerfirnwijfoerewfnfkncheiefowenfcnojrkqepofnkecbjbfiwhefofjgoelghfskjfgjkasgrkeargailrgalugjaeiuryhailfggbaksjdbcdjshkrfelkrh";
        // while(1){
        //     func();
            
        // }
        // write_file(filename, data);
        // read_file(filename);
        // lsdir(disk_mount_pt);

    } else {
        printk("Error mounting disk.\n");
    }


}

void main(void) {

    nrfx_err_t ret;
    sd_card_init();

    IRQ_DIRECT_CONNECT(PDM0_IRQn, 6, pdm_isr_handler, 0);  // workaround line 

    pdm_init();
    

    ret = nrfx_pdm_start();
    if (ret == NRFX_SUCCESS) {
        printk("Pdm start was successful\n");
    } else {
        printk("Pdm start was NOT successful, error code: %d\n", ret);
    }

    while(1){
        write_file(filename, pcm_buf);
        k_sleep(K_SECONDS(3));
    }
    

}

static int lsdir(const char *path)
{
    int res;
    struct fs_dir_t dirp;
    static struct fs_dirent entry;

    fs_dir_t_init(&dirp);								//Initializes the fs_dir_t object; the function needs to be invoked on object before first use with fs_opendir.

    /* Verify fs_opendir() */
    res = fs_opendir(&dirp, path);						//Opens an existing directory specified by the path.
    if (res) {
        printk("Error opening dir %s [%d]\n", path, res);
        return res;
    }

    printk("\nListing dir %s ...\n", path);
    for (;;) {
        /* Verify fs_readdir() */
        res = fs_readdir(&dirp, &entry);					//Reads directory entries of an open directory

        /* entry.name[0] == 0 means end-of-dir */
        if (res || entry.name[0] == 0) {
            break;
        }

        if (entry.type == FS_DIR_ENTRY_DIR) {
            printk("[DIR ] %s\n", entry.name);
        } else {
            printk("[FILE] %s (size = %zu)\n",
                entry.name, entry.size);
        }
    }

    /* Verify fs_closedir() */
    fs_closedir(&dirp);										//Closes an open directory.

    return res;
}

static int write_file(const char *filename, const int16_t *data)
{
    struct fs_file_t file;
    fs_file_t_init(&file);									//Initializes the fs_file_t object; the function needs to be invoked on object before first use with fs_open.

    int res = fs_open(&file, filename, FS_O_CREATE | FS_O_WRITE | FS_O_APPEND);			//Opens or possibly creates a file and associates a stream with it.
    if (res) {
        printk("Error opening file %s [%d]\n", filename, res);
        return res;
    }
    size_t data_len = sizeof(data);
    size_t bytes_written = fs_write(&file, data, data_len);	

    if (bytes_written < data_len) {
        printk("Error writing to file %s [%d]\n", filename, bytes_written);
        fs_close(&file);
        return -EIO;
    }
    else{
        LOG_INF("Wrote %zu bytes to file %s\n", data_len, filename);
    }

    fs_close(&file);													//Flushes the associated stream and closes the file.
    return 0;
}

